package simulator;

import java.util.LinkedList;
import java.util.Random;
import java.util.ArrayList;

public class Network {
	public enum Heuristics {
		RandomSelection, AlmostTriangle, MostNeighborsInO_G, MaxDegree
	};
	
	public enum ThresholdType {
		FixedThreshold, UniformDistribution
	};
	
	private int n; // The number of agents in the network
	private int nA; // The number of agents in community A
	private double p[][]; // The probability of a link [2][2]
	private int h[]; // The threshold for activating a new node
	private int thresholdGen;
	private boolean needGenerate;
	private int linkedNet[][]; // The same networked but in a list-of-neighbors representation
	private boolean activated[];
	private int neighbSum[];
	ThresholdType tp;
	/**
	 * Constructs a Network object with given initial parameters
	 * @param N$ The number of nodes in the graph
	 * @param Na$ The number of nodes in community A
	 * @param Paa$ The probability of the existence of an edge from A to A
	 * @param Pab$ The probability of the existence of an edge from A to B
	 * @param Pba$ The probability of the existence of an edge from B to A
	 * @param Pbb$ The probability of the existence of an edge from B to B
	 * @param H$ The threshold value
	 */
	public Network(int N$, int Na$, double Paa$, double Pab$, double Pba$, double Pbb$, int H$, ThresholdType tp$)
	{
		n = N$;
		nA = Na$;
		p = new double [2][2];
		p[0][0] = Paa$;
		p[0][1] = Pab$;
		p[1][0] = Pba$;
		p[1][1] = Pbb$;
		thresholdGen = H$;
		needGenerate = true;
		activated = new boolean[n];
		neighbSum = new int[n];
		linkedNet = new int[n][];
		h = new int[n];
		tp = tp$;
	}
	
	/**
	 * Generates the actual graph of the network using the parameters received 
	 * by the constructor.
	 */
	public void generate()
	{
		needGenerate = false;
		ArrayList[] Links = new ArrayList[n];
		for (int i = 0; i < n; i++)
		{
			Links[i] =  new ArrayList<Integer>();
		}
		
		for (int i = 0; i < n; i++)
			for (int j = i + 1; j < n; j++)
				if (Math.random() < p[i >= nA ? 1 : 0][j >= nA ? 1 : 0])
				{
					Links[i].add(j);
					Links[j].add(i);
				}
		for (int i = 0; i < n; i++) {
			int num = Links[i].size();
			if (num == 0)
			{
				System.err.println("There are singletons");
			}
			else
			{
				linkedNet[i] = new int [num];
				for (int k = 0; k < num; k++)
					linkedNet[i][k] = Integer.valueOf((Integer) Links[i].get(k));
			}
		}
		Random rnd = new Random();
		for (int i = 0; i < n; i++)
			if (tp == ThresholdType.FixedThreshold)
				h[i] = thresholdGen;
			else
				h[i] = rnd.nextInt(thresholdGen + 1);
	}
	/**
	 * Runs the greedy algorithm
	 * @param nInitialNodes The number of the initially active edges
	 * @param onlyInA If true, runs the greedy algorithm taking as initial nodes
	 * only nodes from the community A
	 * @return The number of finally active nodes after the addition of each 
	 * new initially active node.
	 */
	public int[] simulate(int nInitialNodes, boolean onlyInA, Heuristics heur)
	{
		if (needGenerate)
			generate();
		java.util.Arrays.fill(activated, false);
		java.util.Arrays.fill(neighbSum, 0);
		int answer[] = new int[nInitialNodes + 1];
		answer[0] = 0;
		System.out.println("******** onlyInA = " + (new Boolean(onlyInA)));
		for (int nSeeds = 1; nSeeds <= nInitialNodes; nSeeds++)
		{
			LinkedList<Integer> candidates = new LinkedList<Integer>();
			int maxPropagation = 0;
			int limit = onlyInA ? nA : n;
			for (int v = 0; v < limit; v++)
			{
				if (!activated[v])
				{
					int curPropagation = activate(v, false);
					if (curPropagation == maxPropagation)
					{
						candidates.addLast(v);
					}
					else if (curPropagation > maxPropagation)
					{
						maxPropagation = curPropagation;
						candidates.clear();
						candidates.addLast(v);
					}
				}
				if (v + 1 == limit && maxPropagation == 0)
					limit = n;
			}
			if (!candidates.isEmpty())
			{
				int selection = selectByHeuristics(candidates, heur);
				activate(selection, true);
			}
			answer[nSeeds] = maxPropagation + answer[nSeeds - 1];
		}
		return answer;
	}
	/**
	 * This function receives the Heuristics to use as a parameter, and chooses
	 * an O_G candidate from a given list according to the given Heuristics.
	 * @param candidates The List of possible candidates
	 * @param h The heuristics to use
	 * @return The selected node to be added into O_G
	 */
	private int selectByHeuristics(LinkedList<Integer> candidates, Heuristics h)
	{
		if (h == Heuristics.RandomSelection)
		{
			Random r = new Random();
			return candidates.get(r.nextInt(candidates.size()));
		}
		else if (h == Heuristics.AlmostTriangle)
		{
			int maxTotal = -1;
			LinkedList<Integer> best = new LinkedList<Integer>();
			for (int v : candidates)
			{
				int total = 0;
				for (int i = 0; i < n; i++)
					if (activated[i])
						for (int j = 0, k = 0; j < linkedNet[v].length && k < linkedNet[i].length;)
						{
							if (linkedNet[v][j] < linkedNet[i][k])
								j++;
							else if (linkedNet[v][j] > linkedNet[i][k])
								k++;
							else 
							{
								j++;
								k++;
								total++;
							}
						}
				if (total > maxTotal)
				{
					maxTotal = total;
					best.clear();
					best.addLast(v);
				}
				else if (total == maxTotal)
					best.addLast(v);
			}
			return selectByHeuristics(best, Heuristics.RandomSelection);
		}
		else 
		if (h == Heuristics.MostNeighborsInO_G)
		{
			int maxTotal = -1;
			LinkedList<Integer> best = new LinkedList<Integer>();
			for (int v : candidates)
			{
				int total = 0;
				for (int i = 0; i < linkedNet[v].length; i++)
					if (activated[linkedNet[v][i]])
						total ++;
				if (total > maxTotal)
				{
					maxTotal = total;
					best.clear();
					best.addLast(v);
				}
				else if (total == maxTotal)
					best.addLast(v);
			}
			return selectByHeuristics(best, Heuristics.RandomSelection);
		}
		else if (h == Heuristics.MaxDegree)
		{
			int maxTotal = -1;
			LinkedList<Integer> best = new LinkedList<Integer>();
			for (int v : candidates)
			{
				int total = 0;
				for (int i = 0; i < linkedNet[v].length; i++)
					if (!activated[linkedNet[v][i]])
						total ++;
				if (total > maxTotal)
				{
					maxTotal = total;
					best.clear();
					best.addLast(v);
				}
				else if (total == maxTotal)
					best.addLast(v);
			}
			return selectByHeuristics(best, Heuristics.RandomSelection);
		}
		return -1;
	}
	/**
	 * Activates a given node v
	 * @param v The node to activate
	 * @param persistent Shows if the changes to neighbSum should be persistent
	 * @return The number of edges activated after the activation of v
	 */
	private int activate(int start, boolean persistent)
	{
		LinkedList<Integer> q = new LinkedList<Integer>();
		LinkedList<Integer> all = new LinkedList<Integer>();
		q.push(start);
		while (!q.isEmpty())
		{
			int v = q.removeLast();
			all.push(v);
			for (int i = 0; i < linkedNet[v].length; i++)
			{
				neighbSum[linkedNet[v][i]]++;
				if (neighbSum[linkedNet[v][i]] == h[i])
					q.push(linkedNet[v][i]);
			}
		}
		int count = 0;
		for (int i = 0; i < n; i++)
			if (!activated[i] && (i == start || neighbSum[i] >= h[i]))
			{
				count++;
				if (persistent)
					activated[i] = true;
			}
		if (!persistent)
		{
			for (int v : all)
			{
				for (int i = 0; i < linkedNet[v].length; i++)
					neighbSum[linkedNet[v][i]]--;
			}
		}
		return count;
	}
}
 